Base.@kwdef struct Discretization{
                      T_Grid<:AbstractGrid,
                      T_Cache<:AbstractCache,
                      T_Equation<:AbstractEquation,
                      T_HRSC<:Maybe{AbstractHRSC}}

  grid::T_Grid
  cache::T_Cache
  equation::T_Equation
  hrsc::T_HRSC = nothing
end


struct ParameterDict <: AbstractDict
  provided::FlatDict{String,Any}
  consumed::FlatDict{String,Any}
  defaulted::FlatDict{String,Any}

  function ParameterDict(d::AbstractDict)
    flat_d = FlatDict(d)
    consumed = FlatDict()
    return new(flat_d, consumed)
  end
end


function Base.getindex(pdb::ParameterDict, key)
  val = get(pdb.provided, key, nothing)
  if !isnothing(val)
    if !haskey(pdb.consumed, val)
      pdb.consumed[key] = val
    end
    return val
  end
  throw()
end


function Base.get(pdb::ParameterDict, key, default)
end


function summarize(pdb::ParameterDict)
  # compare pdb.provided, pdb.consumed, pdb.defaulted and display which
  # parameters were unused, which ones were consumed and which ones
  # falled back to defaults
  # the output should also be logable to a file
end


# check if d.provided has key.
# if it has one, copy it over to d.consumed and return it, if not copy the default
# over to d.consumed and return it
Base.get(d::ParameterDict, key, default) = get(d.provided, key, default)
Base.length(d::ParameterDict) = length(d.provided)
# similar to get but throw if key is not found in d.provided
Base.getindex(d::ParameterDict, elements...) = getindex(d.dict, elements...)
# disable this?
Base.setindex!(d::ParameterDict, value, elements...) = setindex!(d.dict, value, elements...)
# iteration interface 
Base.iterate(d::ParameterDict) = iterate(d.dict)
Base.iterate(d::ParameterDict, index) = iterate(d.dict, index)
# can this work with arbitrary dicts?
Base.merge(d::ParameterDict, others::AbstractDict...) = ParameterDict(merge(d.dict, others...))
# disable if we disable setindex!
Base.pop!(d::ParameterDict, key, default) = pop!(d.dict, key, default)






#######################################################################
#                                main                                 #
#######################################################################


parameters_db = ParameterDict(parameters)

# factories, provided by dg1d, except equation should be provided by a project
grid     = make_Grid(parameters_db)
cache    = make_Cache(grid)
hrsc     = make_HRSC(parameters_db)
equation = make_Equation(parameters_db)

# we need an aggregation type
disc = Discretization(grid, cache, hrsc, equation)

# above might be combined into
# and a project must then only implement make_Equation
disc = Discretization(parameters_db)

# prepare initial data, boundary conditions
# this should also be implemented by a project
# and just like in Makie, this should be merely are elaborated expression of what
# problem we want to solve, e.g. which type of initial data we want and what boundary conditions
# no actual computations should occur here
problem = Problem(disc, parameters_db)

# evolution
runner = make_runner(disc, problem, parameters_db)

summarize(runner, parameters_db)
run(runner)

function make_Runner(disc, problem, parameters_db)
  # prepare evolution
  rhs       = make_rhs(disc, problem, parameters_db)
  timestep  = make_timestep(disc, problem, parameters_db)
  callbacks = make_callbacks(disc, problem, parameters_db)
  # () -> dg1d.evolve(rhs,
end


function make_rhs(disc, problem, parameters_db)
  (u, t) -> dg1d.rhs(u, t, disc, problem, disc.equation)
end


function make_timestep(disc, problem, parameters_db)
  (u, t) -> dg1d.timestep(u, t, disc, problem, disc.equation)
end


function dg1d.rhs(u, t, disc, problem, equation::ScalarEquation)
  rhs(u, t, disc, problem, disc.hrsc)
end


function make_callbacks(disc, problem, parameters_db)
  # setup rhs_callbacks, plot_callback, save_callback,
  # id_smoother_callback, aligned_savecallback, restep_callback
end


# what needs to be implemented by a project will be
# (I think we can't use factories because we cannot dispatch on any parameters then, or?)
# - dg1d.make_equation
# - dg1d.make_probelm
# - dg1d.make_rhs
# - dg1d.make_timestep
# - dg1d.make_callback

# things to figure out
# - how to register variables consistently and in such a way that different
# components can depend on one another? do we need another accepts/returns mechanism?
