#######################################################################
#                           Type definition                           #
#######################################################################


struct Grid{N_Dim}
	tree::Tree{N_Dim}
	tensorbasis_cache::ObjectCache{Int64}
	# interpolation_cache::ObjectCache{Int64}
	bases::Vector{TensorBasis{N_Dim}}
	ioffset::Vector{Int64}
	ioffset_boundary::Vector{Int64}
	boxes::Vector{Box{N_Dim}}
	# parameters::Dict{String,Any}

	function Grid{N_Dim}(ns, ps, extends, periodic, blueprint_basis) where N_Dim

		tree = Tree{N_Dim}(ns..., periodic=periodic)
		boxes = place_boxes(tree, extends...)

		tensorbasis_cache = ObjectCache{Int64,SpectralElement}(blueprint_basis)
		bases = [ TensorBasis([ tensorbasis_cache[p] for p in ps]...) for _ in 1:n_cells(tree) ]
		ioffset, ioffset_boundary = Int64[1], Int64[1]
		for b in bases
			push!(ioffset,					ioffset[end]+n_points(b))
			push!(ioffset_boundary, ioffset_boundary[end]+n_points_boundary(b))
		end

		grid = new{N_Dim}(tree, tensorbasis_cache, bases, ioffset, ioffset_boundary, boxes)

		return grid
	end
end
const Grid1d = Grid{Cell1d}
const Grid2d = Grid{Cell2d}
const Grid3d = Grid{Cell3d}


#######################################################################
#                             Constructor                             #
#######################################################################


function Grid(parameters=Dict())
	ns = query(Vector{Int64}, parameters, "ns", default=[5,5,5])
	ps = query(Vector{Int64}, parameters, "ps", default=[ 5 for _ = 1:length(ns) ])
	@assert all(ns .> 0)
	@assert all(ps .== ps[1]) # equal polynomial orders in each direction for now
	N_Dim, n_ps = length(ns), length(ps)
	@assert N_Dim == n_ps """
		Unable to determine dimension of grid. Please provide the same number of parameters for
		- cells per dimension 'ns'
		- polynomial orders per dimensions 'ps'
	"""

	extends = query(Vector{Float64}, parameters, "extends", default=repeat([-1.0,1.0], N_Dim))
	@assert length(extends) == 2*N_Dim """
		Cannot determine domain size from 'extends' = $(extends).
		Require one pair of extrema per dimension.
	"""
	periodic = query(Vector{Bool}, parameters, "periodic", default=[ true for _ in 1:N_Dim ])
	quadrature = query(String, parameters, "quadrature", default="LGL")

	coordinates = query(String, parameters, "domaintype", default="cartesian")

	blueprint(n) = SpectralElement(n, Symbol(quadrature))

	return Grid{N_Dim}(ns, ps, (extends[2*(i-1)+1:2*(i-1)+2] for i = 1:N_Dim), tuple(periodic...), blueprint)
end


#######################################################################
#                              Utilities                              #
#######################################################################


n_points(g::Grid) = sum(n_points.(g.tensorbasis))


function Base.show(io::IO, grid::Grid)
	ncs = ncells(grid)
	npts = npoints(grid)
	nbdrypts = npoints_boundary(grid)
	print("""
				$(typeof(grid))
					# cells                $(ncells(grid))
					# points               $(npoints(grid))
					# boundary points      $(npoints_boundary(grid))
				""")
end



function sample_coordinates(grid::Grid{1})
	xs   = Float64[]
	Js   = Float64[]
	detJ = Float64[]
	npts = n_points(grid)
	sizehint!(xs, npts)
	sizehint!(Js, npts)
	sizehint!(detJ, npts)
	for (box, basis) in zip(grid.boxes, grid.tensorbasis)
		(xmin, xmax), = extrema(basis)
		map_x(z) = (xmax-xmin)/2 * z + (xmax+xmin)/2
		box_xs = [ map_x(zi) for z in basis[Cart1d.X].z ]
		box_Js = (xmax-xmin)/2 * ones(n_points(basis))
		box_detJ = (xmax-xmin)/2
		append!(xs, box_xs)
		append!(Js, box_Js)
		append!(detJ, box_detJ)
	end
	return [ xs ], [ Js ], detJ
end


function sample_coordinates(grid::Grid{2})
	xs, ys = Float64[], Float64[]
	Js_Xx, Js_Xy, Js_Yx, Js_Yy = [ Float64[] for _ = 1:4 ]
	detJ = Float64
	npts = n_points(grid)
	sizehint!(xs, npts)
	sizehint!(ys, npts)
	sizehint!(Js_Xx, npts)
	sizehint!(Js_Xy, npts)
	sizehint!(Js_Yx, npts)
	sizehint!(Js_Yy, npts)
	sizehint!(detJ, npts)
	for (box, basis) in zip(grid.boxes, grid.tensorbasis)
		npts = n_points(basis)
		(xmin, xmax), (ymin, ymax) = extrema(basis)
		map_x(z) = (xmax-xmin)/2 * z + (xmax+xmin)/2
		map_y(z) = (ymax-ymin)/2 * z + (ymax+ymin)/2
		box_xs = [ map_x(zi) for z in basis[Cart2d.X].z ]
		box_ys = [ map_y(zi) for z in basis[Cart2d.Y].z ]
		box_xs, box_ys = mesh_coordinates(box_xs, box_ys)
		box_Js_Xx = (xmax-xmin)/2 * ones(npts)
		box_Js_Xy = zeros(npts)
		box_Js_Yx = zeros(npts)
		box_Js_Yy = (ymax-ymin)/2 * ones(npts)
		box_detJ = @. box_Js_Xx * box_Js_Yy - box_Js_Xy * box_Js_Yx
		push!(xs, box_xs)
		push!(ys, box_ys)
		push!(Js_Xx, box_Js_Xx)
		push!(Js_Xy, box_Js_Xy)
		push!(Js_Yx, box_Js_Yx)
		push!(Js_Yy, box_Js_Yy)
		push!(detJ, box_detJ)
	end
	return [ xs, ys ], [ Js_Xx, Js_Xy, Js_Yx, Js_Yy ], detJ
end


function sample_coordinates(grid::Grid{3})
	xs, ys, zs = Float64[], Float64[], Float64[]
	Js_Xx, Js_Xy, Js_Xz, Js_Yx, Js_Yy, Js_Yz, Js_Zx, Js_Zy, Js_Zz = [ Float64[] for _ = 1:9 ]
	detJ = Float64[]
	npts = n_points(grid)
	sizehint!(xs, npts)
	sizehint!(ys, npts)
	sizehint!(zs, npts)
	sizehint!(Js_Xx, npts)
	sizehint!(Js_Xy, npts)
	sizehint!(Js_Xz, npts)
	sizehint!(Js_Yx, npts)
	sizehint!(Js_Yy, npts)
	sizehint!(Js_Yz, npts)
	sizehint!(Js_Zx, npts)
	sizehint!(Js_Zy, npts)
	sizehint!(Js_Zz, npts)
	sizehint!(detJ, npts)
	for (box, basis) in zip(grid.boxes, grid.tensorbasis)
		npts = n_points(basis)
		(xmin, xmax), (ymin, ymax), (zmin, zmax) = extrema(basis)
		map_x(z) = (xmax-xmin)/2 * z + (xmax+xmin)/2
		map_y(z) = (ymax-ymin)/2 * z + (ymax+ymin)/2
		map_z(z) = (zmax-zmin)/2 * z + (zmax+zmin)/2
		box_xs = [ map_x(zi) for z in basis[Cart3d.X].z ]
		box_ys = [ map_y(zi) for z in basis[Cart3d.Y].z ]
		box_zs = [ map_z(zi) for z in basis[Cart3d.Z].z ]
		box_xs, box_ys, box_zs = mesh(box_xs, box_ys, box_zs)
		box_Js_Xx = (xmax-xmin)/2 * ones(npts)
		box_Js_Xy = zeros(npts)
		box_Js_Xz = zeros(npts)
		box_Js_Yx = zeros(npts)
		box_Js_Yy = (ymax-ymin)/2 * ones(npts)
		box_Js_Yz = zeros(npts)
		box_Js_Zx = zeros(npts)
		box_Js_Zy = zeros(npts)
		box_Js_Zz = (zmax-zmin)/2 * ones(npts)
		box_detJ = ( @. box_Js_Xx * box_Js_Yy * box_Zz + box_Js_Xy * box_Js_Yz * box_Js_Zx +
									  box_Js_Xz * box_Js_Yx * box_Zy - box_Js_Xz * box_Js_Yy * box_Js_Zx -
										box_Js_Xy * box_Js_Yx * box_Zz - box_Js_Xx * box_Js_Yz * box_Js_Zy )
		append!(xs, box_xs)
		append!(ys, box_ys)
		append!(zs, box_zs)
		push!(Js_Xx, box_Js_Xx)
		push!(Js_Xy, box_Js_Xy)
		push!(Js_Xz, box_Js_Xz)
		push!(Js_Yx, box_Js_Yx)
		push!(Js_Yy, box_Js_Yy)
		push!(Js_Yz, box_Js_Yz)
		push!(Js_Zx, box_Js_Zx)
		push!(Js_Zy, box_Js_Zy)
		push!(Js_Zz, box_Js_Zz)
		push!(detJ, box_detJ)
	end
	return [ xs, ys, zs ], [ Js_Xx, Js_Xy, Js_Xz, Js_Yx, Js_Yy, Js_Yz, Js_Zx, Js_Zy, Js_Zz ], detJ
end


function mesh_coordinates(x, y)
	nx, ny = length(x), length(y)
	xx, yy = [ zeros(nx*ny) for _ in 1:2 ]
	for i in 1:nx, j in 1:ny
		ij = i + (j-1) * nx
		xx[ij] = x[i]
		yy[ij] = y[j]
	end
	return xx, yy
end


function mesh_coordinates(x,y,z)
	nx, ny, nz = length(x), length(y), length(z)
	xx, yy, zz = [ zeros(nx*ny*nz) for _ in 1:3 ]
	for i in 1:nx, j in 1:ny, k in 1:nz
		ijk = i + (j-1) * nx + (k-1) * nx * ny
		xx[ijk] = x[i]
		yy[ijk] = y[j]
		zz[ijk] = z[k]
	end
	return xx, yy, zz
end


function register_cartesian_coordinates!(grid::Grid)
	xys, components_Js, detJ = sample_coordinates(grid)
	add_variables!(grid.cache, static_variables = [ xs..., Js..., detJ ])
end


function regrid!(grid::Grid)
	# duplicate tree, bases and boxes and update them
	# regrid cached variables inplace using the above copies
	# e.g. perform various interpolation coarsening steps in different types of variables
end


function eachcell(grid::Grid, vars::RefBlockVector...)
	((block(v, i) for i in 1:n_cells(grid)) for v in vars) 
end


# function range(grid::Grid, i)
#   range(grid.ioffset[i], length=n_points(grid.bases[i]))
# end


# struct Boundary{T<:Grid{N_Dim}}
#   grid::T
#   dir::Cart{N_Dim}.Direction
# end
#
#
# function eachcell(g::Boundary{Grid}, vars...)
#   ((view(v, range(grid, i, dir)) for v in vars) for i in 1:n_cells(grid))
# end
#
#
# function range(cell::Grid, i, dir)
#   range(bdry.grid.ioffset[i]+offset_boundary(grid.bases[i], dir),
#         length=n_points_boundary(grid.bases[i], dir))
# end


# for (ai,bi,ci) in eachcell(grid,a,b,c)
#   ci .= ai .+ bi
# end
# for (ai,bi,ci) in eachcell(Boundary(grid,Cart2d.Xmin),a,b,c)
#   ci .= ai .+ bi
# end
