export VariableCache, register_variables!, add_variables!, wrap_variables!


#######################################################################
#                                Types                                #
#######################################################################


"""
    VariableVariableCache

Caching object to bundle variables of different kinds.
"""
struct VariableCache{T_Grid<:AbstractGrid}
  grid::T_Grid
  varlist::Dict{Symbol,Symbol}
  dynamic::Variables
  static::Variables
  cell::Variables
  rhs::Variables
  # interface::Variables # should we added this to separate inferface and boundary values
  # would allow to easily distinguish boundary values that are computed from boundary conditions
  # and/or neighboring grids (e.g. sphered cubes etc).
  boundary::Variables
  domain::Variables
end


function VariableCache(grid::Grid)
  verbose = false
  varlist        = Dict{Symbol, Symbol}()
  npts, npts_bdry     = n_points(grid), n_points_boundary(grid)
  ncells              = n_cells(grid)
  dynamic             = Variables(npts, verbose=verbose)
  static              = Variables(npts, verbose=verbose)
  cell                = Variables(ncells, verbose=verbose)
  rhs                 = Variables(npts, verbose=verbose)
  boundary            = Variables(npts_bdry, verbose=verbose)
  domain              = Variables(1, verbose=verbose)
  return VariableCache(grid, varlist, dynamic, static, cell, rhs, boundary, domain)
end


#######################################################################
#                               Methods                               #
#######################################################################


@inline UnPack.unpack(x::VariableCache, ::Val{f}) where {f} = get_variable(x, Val(f))


function get_variable(varcache::VariableCache, ::Val{f}) where {f}
  field = v.varlist[f]
  return getproperty(varcache, field)
end




# TODO Can we avoid this and instead overload Unpack.@unpack or so?
function Base.propertynames(::VariableCache)
  fnames = fieldnames(VariableCache)
  filter(f -> !in(f, [ :varlist ]), fnames)
end


"""
    wrap_variables!(C::VariableCache, u=nothing, du=nothing)

Wrap fields `dynamic, rhs` around data `u, du` that is not owned by `C`.
"""
function wrap_variables!(C::VariableCache; u=nothing, du=nothing)
  !isnothing(u) && set!(C.dynamic, u)
  !isnothing(du) && set!(C.rhs, du)
  return
end



function register_variables!(C::VariableCache;
    dynamic = [], rhs = [], static = [], cell = [], bdry = [], domain = [])

  register!(C.dynamic,  Symbol.(dynamic)...)
  register!(C.rhs,      Symbol.(rhs)...)
  register!(C.static,   Symbol.(static)...)
  register!(C.cell,     Symbol.(cell)...)
  register!(C.boundary, Symbol.(boundary)...)
  register!(C.domain,   Symbol.(domain)...)

  empty!(cache.varlist)
  push_names_field!(cache.varlist, dg1d.names(cache.dynamic), :dynamic)
  push_names_field!(cache.varlist, dg1d.names(cache.rhs), :rhs)
  push_names_field!(cache.varlist, dg1d.names(cache.static), :static)
  push_names_field!(cache.varlist, dg1d.names(cache.cell), :cell)
  push_names_field!(cache.varlist, dg1d.names(cache.boundary), :boundary)
  push_names_field!(cache.varlist, dg1d.names(cache.domain), :domain)

  return
end


push_names_field!(dict, names, field) = for name in names; push!(dict, name => field); end


function Base.show(io::IO, C::VariableCache)
  variables_list = collect(keys(C.fields_to_variables))
  N_variables = length(variables_list)
  if N_variables == 0 
    @warn "No variables registered!"
    return
  end
  names_list = collect(values(C.fields_to_variables))
  permute_idx = sortperm(variables_list)
  variables_list, names_list = variables_list[permute_idx], names_list[permute_idx]
  max_rows = maximum(length.(names_list))
  data = Matrix{String}(undef, max_rows, N_variables)
  fill!(data, "")
  for (idx, names) in enumerate(names_list)
    N_names = length(names)
    data[1:N_names,idx] .= string.(names)
  end
  # sort!(data, dims=1)
  pretty_table(data, header=variables_list, crop=:none)
end


function resize!(varcache::VariableCache)
  npts = n_points(varcache)
  npts_bdry = n_points_boundary(varcache)
  ncells = n_cells(varcache)
  resize!(varcache.dynamic)
  resize!(varcache.rhs, npts)
  resize!(varcache.static, npts)
  resize!(varcache.cell, ncells)
  resize!(varcache.boundary, npts_bdry)
  # domain unchanged, because it only contains single values
end


function regrid_dynamic!(varcache::VariableCache)
  for name in varcache.dynamic
    
  end
end
