struct Grid{N_Dim,T_SpectralElement}
  mesh::TreeMesh{N_Dim}
  bases_cache::ObjectCache{Int64,T_SpectralElement}
  bases::Vector{TensorBasis{N_Dim}}
  cache::Cache
end


@unpack variable"u, v, w, fu, fv, fw" = cache
@unpack dynamic"u, v, w" = cache
@unpack static"u, v, w" = cache
@unpack boundary"u, v, w" = cache
@unpack cell"u, v, w" = cache
# register_variable.(cache, dynamic"u, v, w")
->
getproperty(getproperty(cache, :dynamic), :u)
(alternative) ->
getproperty(cache, Dynamic(:u))
(intermediate) ->
@unpack Var(:u), Var(:v), Var(:w), ... = cache


function Grid(tm::TreeMesh{N_Dim},
    grid_extends::Vararg{Tuple{Float64,Float64},N_Dim}) where N_Dim

  ncells = n_cells(tm)
  bases = [ TensorBasis(plevels(c)) for c in cells(tm) ]
end



const Grid1d = Grid{Cell1d}
const Grid2d = Grid{Cell2d}
const Grid3d = Grid{Cell3d}


function Grid(ncells)
  grid = Grid1d([ Cell1d() for _ in 1:ncells ])
  @unpack cells = grid
  # connect cells
  for (i, cell) in enumerate(cells)
    i_left, i_right = periodize(i-1, ncells), periodize(i+1, ncells)
    neighbours = Neighbours1d( CellId(i_left, VolumeCell), CellId(i_right, VolumeCell) )
    setproperty!(cell, :neighbours, neighbours)
    setproperty!(cell, :this, i)
  end
  return grid
end


function Grid(ncells_x, ncells_y)
  ncells = ncells_x * ncells_y
  grid = Grid2d([ Cell2d() for _ in 1:ncells ])
  @unpack cells = grid
  CI = CartesianIndices((ncells_x, ncells_y))
  for (i, cell) in enumerate(cells)
    ix, iy = Tuple(CI[i])
    i_left, i_right = periodize(ix-1, ncells_x), periodize(ix+1, ncells_x)
    i_top, i_bottom = periodize(iy-1, ncells_y), periodize(iy+1, ncells_y)
    neighbours = Neighbours2d( CellId(i_left, VolumeCell), CellId(i_right, VolumeCell),
                               CellId(i_top, VolumeCell), CellId(i_bottom, VolumeCell) )
    setproperty!(cell, :neighbours, neighbours)
    setproperty!(cell, :this, i)
  end
  return grid
end


function Grid(ncells_x, ncells_y, ncells_z)
  ncells = ncells_x * ncells_y * ncells_z
  grid = Grid3d([ Cell3d() for _ in 1:ncells ])
  @unpack cells = grid
  CI = CartesianIndices((ncells_x, ncells_y, ncells_z))
  for (i, cell) in enumerate(cells)
    ix, iy, iz = Tuple(CI[i])
    i_left, i_right = periodize(ix-1, ncells_x), periodize(ix+1, ncells_x)
    i_top, i_bottom = periodize(iy-1, ncells_y), periodize(iy+1, ncells_y)
    i_front, i_back = periodize(iz-1, ncells_z), periodize(iz+1, ncells_z)
    neighbours = Neighbours3d( CellId(i_left, VolumeCell), CellId(i_right, VolumeCell),
                               CellId(i_top, VolumeCell), CellId(i_bottom, VolumeCell),
                               CellId(i_front, VolumeCell), CellId(i_back, VolumeCell) )
    setproperty!(cell, :neighbours, neighbours)
    setproperty!(cell, :this, i)
  end
  return grid
end


cells(grid::Grid) = grid.cells
Base.ndims(grid::Grid) = ndims(eltype(cells(grid)))
ncells(grid::Grid) = length(cells(grid))
npoints(grid::Grid) = sum(npoints(c) for c in cells(grid))
npoints_boundary(grid::Grid1d) = 2
npoints_boundary(grid::Grid) = sum(npoints_boundary(c) for c in cells(grid))
nneighbours(grid::Grid) = sum(nneighbours(c) for c in cells(grid))


const _CACHE_SpectralElements = Dict{Int64,SpectralElement}()


function make_SpectralElement(; args...)
  N = args[:N]
  if haskey(_CACHE_SpectralElements, N)
    return _CACHE_SpectralElements[N]
  end
  el = SpectralElement(values(args)...)
  _CACHE_SpectralElements[N] = el
  return el
end


function Base.fill!(grid::Grid, basis::Basis)
  @assert ndims(basis) == ndims(grid)
  npts = npoints(basis)
  offset = 1
  for cell in cells(grid)
    setproperty!(cell, :basis, basis)
    setproperty!(cell, :offset, offset)
    offset += npts
  end
end


function Base.show(io::IO, grid::Grid)
  ncs = ncells(grid)
  npts = npoints(grid)
  nbdrypts = npoints_boundary(grid)
  print("""
        $(typeof(grid))
          # cells                $(ncells(grid))
          # points               $(npoints(grid))
          # boundary points      $(npoints_boundary(grid))
        """)
end

# TODO Show methods for cell and basis.
