export Mesh, make_Mesh


#######################################################################
#                                Types                                #
#######################################################################


abstract type AbstractMesh{Ndim} end

@doc """
    Mesh{N}

Object that bundles grid and operators needed for modeling a Discontinuous Galerkin 
scheme in N dimensions.
"""
struct Mesh{Ndim} <: AbstractMesh{Ndim}
  elements::NTuple{Ndim,SpectralElement} # collection of spectral element operators
  Ks::NTuple{Ndim,Int64}                 # number of cells
  bounds::NTuple{Ndim,NTuple{Ndim,Float64}}
  CFL::Float64                           # Courant Friedrichs Lewy factor
  Ls::NTuple{Ndim,Float64}               # domain length
  dls::NTuple{Ndim,Float64}              # equidistant subcell length
  xs::NTuple{Ndim,Matrix{Float64}}       # grid points
  invjacs::NTuple{Ndim,Matrix{Float64}}  # inverse jacobian of coordinate transformation
                                         # from computational [-1,1] onto physical grid [xmin,xmax]

  function Mesh{Ndim}(; Ns, Ks, bounds, CFL, quadrature_method)
    elements = [ SpectralElement(N, Symbol(quadrature_method)) for N in Ns ]
    zs       = [ el.z for el in elements ]
    bounds   = [ extrema(el.z) for el in elements ]
    Ls       = [ diff(b) for b in bounds ]
    dls      = [ L/K for (L,K) in zip(Ls, Ks) ]
    as       = [ [ b[1] + dl * idx for idx = 0 : K-1 ] for (b,dl,K) in zip(bounds, Ls, Ks) ]
    bs       = [ a.+dl for (a,dl) in zip(as,dls) ]
    Npts     = [ length(z) for z in zs ]
    xs       = [ Matrix{Float64}(undef, Np, K) for (Np,K) in zip(Npts, Ks) ]
    invjacs  = [ similar(x) for x in xs ]
    for (Np, K, a, b, z, x, invjac) in zip (Npts, Ks, as, bs, zs, xs, invjacs)
      for i=1:Np, j=1:K
        aj = a[j]
        bj = b[j]
        zi = z[i]
        x[i,j] = 0.5 * ((bj - aj) * zi + (aj + bj))
        invjac[i,j] = 2.0 / (bj - aj)
      end
      # sanitize cell end points to have same x coordinate at common interfaces
      for j=1:K-1
        xl = x[end,j]
        xr = x[1,j+1]
        xmean = 0.5 * (xl + xr)
        x[end,j] = xmean
        x[1,j+1] = xmean
      end
    end
    return new{Ndim}(elements, Tuple(Ks), bounds, CFL, Ls, dls, xs, invjacs)
  end
end


#######################################################################
#                              Factories                              #
#######################################################################


make_Mesh(prms)      = Mesh(; parse_parameters(Mesh, prms)...)
make_Mesh(; args...) = make_Mesh(FlatDict(args))


######################################################################
#                            Parameters                              #
######################################################################


const Mesh1d = Mesh{1}
const Mesh2d = Mesh{2}
const Mesh3d = Mesh{3}


function parse_parameters(::Type{Mesh{Ndim}}, prms)
  N = query(Vector{Int64}, prms, "N", default=repeat([5],Ndim))
  K = query(Vector{Int64}, prms, "K", default=repeat([10],Ndim))
  CFL = query(Float64, prms, "CFL")
  quadrature_method = query(String, prms, "quadrature_method", default="LGL",
                            options=["LGL", "GLGL", "LGL_lumped"])
  range = query(Vector{Vector{Float64}}, prms, "range", default=repeat([[-1.0,1.0]], Ndim))
  @assert all(N .> 0) "'N > 0' required"
  @assert all(K .> 0) "'K > 0' required"
  @assert CFL > 0 "'CFL > 0' required"
  @assert all(length.(range) .== 2) "'range' must be a tuple"
  # @assert all(range[1] < xrange[2] "'xrange' limits must be ordered"
  @assert quadrature_method in ["LGL", "GLGL", "LGL_lumped"] "'quadrature_method' invalid"
  return N, K, CFL, quadrature_method, range
end


function parse_parameters(::Type{Mesh1d}, prms)
  N, K, CFL, quadrature_method, range = parse_parameters(Mesh, prms)


end


function parse_parameters(::Type{Mesh1d}, prms)
  Ns = query(Int64, prms, "N", default=[5])
  Ks = query(Int64, prms, "K", default=[20])
  CFL = query(Float64, prms, "CFL", default=0.5)
  quadrature_method = query(String, prms, "quadrature_method", default="LGL",
                            options=["LGL", "GLGL", "LGL_lumped"])
  xrange = query(Vector{Float64}, prms, "xrange", default = [-1.0,1.0])
  # sanitize
  @assert Ns > 0 "'N > 0' required"
  @assert Ks > 0 "'K > 0' required"
  @assert CFL > 0 "'CFL > 0' required"
  @assert length(xrange) == 2 "'xrange' must be a tuple"
  @assert xrange[1] < xrange[2] "'xrange' limits must be ordered"
  @assert quadrature_method in ["LGL", "GLGL", "LGL_lumped"] "'quadrature_method' invalid"
  bounds = (xrange,)
  return (; Ns, Ks, CFL, quadrature_method, bounds)
end


function parse_parameters(::Type{Mesh2d}, prms)
  Ns = query(NTuple{2,Int64}, prms, "N", default=(5,5))
  Ks = query(NTuple{2,Int64}, prms, "K", default=(20,20))
  CFL = query(Float64, prms, "CFL", default=0.5)
  quadrature_method = query(String, prms, "quadrature_method", default="LGL",
                            options=["LGL", "GLGL", "LGL_lumped"])
  xrange = query(Vector{Float64}, prms, "xrange", default = [-1.0,1.0])
  yrange = query(Vector{Float64}, prms, "yrange", default = [-1.0,1.0])
  # sanitize
  @assert all(Ns .> 0) "'N > 0' required"
  @assert all(Ks .> 0) "'K > 0' required"
  @assert CFL > 0 "'CFL > 0' required"
  @assert length(xrange) == 2 "'xrange' must be a tuple"
  @assert length(yrange) == 2 "'yrange' must be a tuple"
  @assert xrange[1] < xrange[2] "'xrange' limits must be ordered"
  @assert yrange[1] < yrange[2] "'yrange' limits must be ordered"
  @assert quadrature_method in ["LGL", "GLGL", "LGL_lumped"] "'quadrature_method' invalid"
  bounds = (xrange, yrange)
  return (; Ns, Ks, CFL, quadrature_method, bounds)
end


function parse_parameters(::Type{Mesh3d}, prms)
  Ns = query(NTuple{3,Int64}, prms, "N", default=(5,5,5))
  Ks = query(NTuple{3,Int64}, prms, "K", default=(20,20,20))
  CFL = query(Float64, prms, "CFL", default=0.5)
  quadrature_method = query(String, prms, "quadrature_method", default="LGL",
                            options=["LGL", "GLGL", "LGL_lumped"])
  xrange = query(Vector{Float64}, prms, "xrange", default = [-1.0,1.0])
  yrange = query(Vector{Float64}, prms, "yrange", default = [-1.0,1.0])
  zrange = query(Vector{Float64}, prms, "zrange", default = [-1.0,1.0])
  # sanitize
  @assert all(Ns .> 0) "'N > 0' required"
  @assert all(Ks .> 0) "'K > 0' required"
  @assert CFL > 0 "'CFL > 0' required"
  @assert length(xrange) == 2 "'xrange' must be a tuple"
  @assert length(yrange) == 2 "'yrange' must be a tuple"
  @assert length(zrange) == 2 "'zrange' must be a tuple"
  @assert xrange[1] < xrange[2] "'xrange' limits must be ordered"
  @assert yrange[1] < yrange[2] "'yrange' limits must be ordered"
  @assert zrange[1] < zrange[2] "'zrange' limits must be ordered"
  @assert quadrature_method in ["LGL", "GLGL", "LGL_lumped"] "'quadrature_method' invalid"
  bounds = (xrange, yrange, zrange)
  return (; Ns, Ks, CFL, quadrature_method, bounds)
end


#######################################################################
#                               Methods                               #
#######################################################################


Base.broadcastable(mesh::AbstractMesh)                  = Ref(mesh)
Base.ndims(mesh::AbstractMesh{N})                       = N
coordinates(mesh::AbstractMesh)                         = mesh.xs
bounds(mesh::AbstractMesh)                              = mesh.bounds
Base.length(mesh::AbstractMesh{N})::NTuple{N,Int64}     = prod(length.(coordinates(mesh)))
n_points(mesh::AbstractMesh{N})                         = mesh.element.Npts .+ 1
n_cells(mesh::AbstractMesh{N})                          = mesh.Ks
layout(mesh::AbstractMesh{N})::NTuple{N,Tuple{Int,Int}} = (n_points(mesh), n_cells(mesh))


"""
    cell_bdrys(mesh::AbstarctMesh)

Return coordinates of cell boundaries from a `mesh`.
"""
function cell_bdrys(mesh::Mesh{N}) where N
  bdrys = NTuple{N,Vector{Float64}}([ Vector{Float64}(undef, K+1) for K in Ks ])
  for (idx,xi) in enumerate(coordinates(mesh))
    bdrys[idx] .= vcat(x[1,:], [x[end,end]])
  end
  return bdrys
end

@deprecate extract_cell_bdrys(mesh) cell_bdrys(mesh)

@doc """
    grid_average(u, mesh::Mesh)

Return the grid average of all cell averages 
"""
function grid_average(u, mesh::Mesh1d)
  @unpack element, dl, L = mesh
  cellavg = cellwise_average(u, mesh) .* dl
  avg = sum(cellavg) / L
  return avg
end


@doc """
  `cellwise_average(u, mesh::Mesh)`

Return the cellwise average of `u` over `mesh.x`.
"""
function cellwise_average(u, mesh::Mesh)
  @unpack element, invjac, K, dl = mesh
  @unpack Npts = element
  v = ones(Npts)
  avg = zeros(K)
  for i in 1:K
    avg[i] = element.iproduct(u[:,i], v)
  end
  avg ./= invjac[1,:] * dl # constant Jacobian, constant cell width
  return avg
end


@doc """
  `cellwise_inner_product(u, v, mesh::Mesh)`

Return vector of inner products between `u, v` on every cell in `mesh`.
"""
function cellwise_inner_product(u, v, mesh::Mesh)
  @unpack element, K, invjac = mesh
  @unpack Npts = element
  L = layout(mesh)
  mat_u = reshape(view(u,:), L)
  mat_v = reshape(view(v,:), L)
  inner_products = [ element.iproduct(mat_u[:,k], mat_v[:,k]) for k = 1:K ]
  inner_products ./= invjac[1,:] # constant Jacobian
  return inner_products
end


@doc """
  `broken_inner_product(u, v, mesh::Mesh)`
  
Return sum of all inner products between `u,v` on every cell in `mesh`.
"""
function broken_inner_product(u, v, mesh::Mesh)
  inner_products = cellwise_inner_product(u, v, mesh)
  total = sum(inner_products[:])
  return total
end


# function LinearAlgebra.norm(u, mesh::Mesh, order)
#   @match order begin
#     1 => norm_L1(u, mesh)
#     2 => norm_L2(u, mesh)
#     _ => error("Norm for order = $order not implemented.")
#   end
# end
# norm_L2(u, mesh::Mesh) = sqrt(broken_inner_product(u, u, mesh))
# norm_L1(u, mesh::Mesh) = broken_inner_product(abs.(u), ones(Float64, size(u)), mesh)


@doc """
Compute local maximum absolute value.
"""
function local_maxabs(v::AbstractMatrix, mesh::Mesh) 
  return vec(maximum(abs.(v), dims=1))
end


@doc """
Compute local minimum absolute value.
"""
function local_maxabs(v::AbstractVector, mesh::Mesh) 
  return local_maxabs(reshape(v, (mesh.element.Npts, mesh.K)))
end


"""
    find_nearest(xi, mesh::Mesh1d)
    find_nearest(xi, mesh::Mesh2d)
    find_nearest(xi, mesh::Mesh3d)

Find index of node in `mesh` that lies closest to `xi`.
Returns the index of the outermost point if `xi` falls outside `mesh`.
"""
function find_nearest(pt_x, mesh::Mesh1d)
  px = pt_x
  x, = coordinates(mesh)
  min_ix = 1
  min_distance = abs(x[min_ix] - px)
  for ix = 2:length(x)
    distance = abs(x[ix] - px)
    if distance < min_distance
      min_distance = distance
      min_ix = ix
    end
  end
  return (min_ix,)
end
function find_nearest(pt_xy, mesh::Mesh2d)
  px, py = pt_xy
  xs, ys = coordinates(mesh)
  min_ix, min_iy = 1, 1
  min_distance = sqrt( (xs[min_ix]-px)^2 + (ys[min_iy]-py)^2 )
  for ix = 2:length(x), iy = 2:length(y)
    distance = sqrt( (xs[ix]-px)^2 + (ys[iy]-py)^2 )
    if distance < min_distance
      min_distance = distance
      min_ix, min_iy = ix, iy
    end
  end
  return (min_ix, min_iy)
end
function find_nearest(pt_xyz, mesh::Mesh2d)
  px, py, pz = pt_xyz
  xs, ys, zs = coordinates(mesh)
  min_ix, min_iy, min_iz = 1, 1, 1
  min_distance = sqrt( (xs[min_ix]-px)^2 + (ys[min_iy]-py)^2 + (zs[min_iz]-pz)^2 )
  for ix = 2:length(x), iy = 2:length(y), iz = 2:length(iz)
    distance = sqrt( (xs[ix]-px)^2 + (ys[iy]-py)^2 + (zs[min_iz]-pz)^2 )
    if distance < min_distance
      min_distance = distance
      min_ix, min_iy, min_iz = ix, iy, iz
    end
  end
  return (min_ix, min_iy, min_iz)
end


@deprecate locate_point(xi, mesh::Mesh) find_nearest(xi, mesh::Mesh)
@deprecate locate_point(xi, mesh::AbstractMesh) find_nearest(xi, mesh::AbstractMesh)


"""
    find_cell_index(xi, mesh::Mesh)

Find the cell index into which coordinate `xi` falls into the `mesh`.
Note that it returns the first index found in case `xi` falls onto an interface boundary.
Returns nothing if `xi` lies outside the `mesh`.
"""
function find_cell_index(pt_x, mesh::Mesh1d)
  px = pt_x
  xbdrys, = cell_bdrys(mesh)
  return findfirst(1:length(xbdrys)-1) do ix
    xbdrys[ix] <= px <= xbdrys[ix+1]
  end
end
function find_cell_index(pt_xy, mesh::Mesh2d)
  px, py = pt_xy
  xbdrys, ybdrys = cell_bdrys(mesh)
  return findfirst((1:length(xbdrys)-1,1:length(ybdrys)-1)) do ix, iy
    xbdrys[ix] <= px <= xbdrys[ix+1] && ybdrys[iy] <= py <= ybdrys[iy+1]
  end
end
function find_cell_index(pt_xyz, mesh::Mesh3d)
  px, py, pz = pt_xyz
  xbdrys, ybdrys, zbdrys = cell_bdrys(mesh)
  return findfirst((1:length(xbdrys)-1,1:length(ybdrys)-1,1:length(zbdrys)-1)) do ix, iy, iz
    xbdrys[ix] <= px <= xbdrys[ix+1] && 
    ybdrys[iy] <= py <= ybdrys[iy+1] && 
    zbdrys[iz] <= pz <= zbdrys[iz+1]
  end
end
