Base.@kwdef struct RHS{T_Equation <:SREulerEquation, 
                       T_RSolver  <:AbstractRiemannSolver,
                       T_HRSC     <:Maybe{HRSC.AbstractHRSC},
                       T_TCI      <:Maybe{TCI.AbstractTCI},
                       T_LDG_RSolver<:Maybe{AbstractRiemannSolver},
                       T_AV_RSolver<:Maybe{AbstractRiemannSolver}}

  cache::Cache
  mesh::Mesh
  equation::T_Equation
  rsolver::T_RSolver
  hrsc::T_HRSC               = nothing
  tci::T_TCI                 = nothing
  ldg_rsolver::T_LDG_RSolver = nothing
  av_rsolver::T_AV_RSolver   = nothing
end


#######################################################################
#                         Register variables                          #
#######################################################################


# Julia limitation/bug: Overloading imported functions without their Module prefix shadows their
# definition in this module without a warning or error. This does not happen for methods
# defined in Base.
# Because of that we prefix with _ below.
# Cf. https://github.com/JuliaLang/julia/issues/15510
function _register_variables!(rhs::RHS)
  register_variables!(rhs.cache,
    dynamic_variablenames = (:D, :S, :tau,),            # conservatives
    rhs_variablenames     = (:rhs_D, :rhs_S, :rhs_tau), # rhs of conservatives
    static_variablenames  = (:flx_D, :flx_S, :flx_tau,  # conservatives' fluxes
                             :rho, :v, :p, :eps,        # primitives
                             :max_v, :x,                # maximum charactersitic speed, grid
                             :E, :Em1,                  # entropy
                             :flx_E, :flx_Em1),         # entropy flux
    cell_variablenames    = (:cellmax_v,),
    global_variablenames  = (:t, :tm1) # time steps
  )
  _register_variables!(rhs, rhs.hrsc)
  _register_variables!(rhs, rhs.tci)
  @unpack x = get_static_variables(rhs.cache)
  @. x = rhs.mesh.x[:]
end


function _register_variables!(rhs, tci_or_hrsc::Nothing) end


_register_variables!(rhs, hrsc::AbstractHRSC) = 
  error("_register_variables: Not implemented for hrsc of type '$(typeof(hrsc))'")

_register_variables!(rhs, hrsc::HRSC.AbstractReconstruction) = nothing


function _register_variables!(rhs, hrsc::HRSC.AbstractArtificialViscosity)
  register_variables!(rhs.cache,
    static_variablenames  = (:ldg_D, :ldg_S, :ldg_tau,              # local DG variable
                             :flx_ldg_D, :flx_ldg_S, :flx_ldg_tau), # local DG flux
    cell_variablenames    = (:mu,), # 'one viscosity to rule them all'
    )
end


function _register_variables!(rhs, hrsc::HRSC.SmoothedArtificialViscosity)
  _register_variables!(rhs, hrsc.av)
  register_variables!(rhs.cache,
    static_variablenames  = (:smoothed_mu,)
  )
end


function _register_variables!(rhs, tci::TCI.AbstractTCI)
  register_variables!(rhs.cache, cell_variablenames = (:flag,:D_flag,:S_flag,:tau_flag))
end


function _register_variables!(rhs, tci::TCI.EntropyProduction)
  register_variables!(rhs.cache, 
                      cell_variablenames = (:flag,),
                      static_variablenames = (:EP,), # entropy production
                      bdry_variablenames   = (:EPjump_lhs, :EPjump_rhs)
                                              # entropy production due to jumps at interfaces
  )
end


#######################################################################
#                              Timestep                               #
#######################################################################


function timestep(state_u, t, rhs::RHS)
  wrap_dynamic_variables!(rhs.cache, state_u)
  return timestep(rhs, rhs.hrsc)
end


function timestep(rhs::RHS, hrsc::Maybe{HRSC.AbstractHRSC})
  @unpack cache, mesh, equation = rhs
  @unpack max_v = get_static_variables(cache)

  broadcast_volume!(cons2prim, equation, cache)
  broadcast_volume!(speed, equation, cache)
  vmax = dg1d.absolute_maximum(max_v)
  # vmax > 0.99 && error("Unreasonable maximum speed encountered: $vmax")
  vmax_limit = 0.9999
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end

  @unpack dl, x, CFL, element = mesh
  @unpack N = element

  dt = CFL * dl / (N^2 * vmax)
  return dt
end


function timestep(rhs::RHS, hrsc::HRSC.AbstractArtificialViscosity)
  @unpack cache, mesh, equation = rhs

  broadcast_volume!(cons2prim, equation, cache)
  broadcast_volume!(speed, equation, cache)
  max_v = get_variable(cache, :max_v)
  vmax = dg1d.absolute_maximum(view(max_v,:))
  vmax_limit = 0.99
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end
  smoothed_mu = get_variable(cache, :smoothed_mu)
  mumax = dg1d.absolute_maximum(view(smoothed_mu, :))

  @unpack dl, x, CFL, element = mesh
  @unpack N = element

  dt = CFL / (vmax * N / dl + mumax * N^3 / dl^2)
  dt_limit = 1e-7
  if dt < dt_limit
    @warn "Limiting timestep due to dt being smaller than $dt_limit"
    dt = dt_limit
  end
  return dt
end


#######################################################################
#                           RHS computation                           #
#######################################################################


function rhs!(state_du, state_u, t, rhs::RHS, bdryconds::BoundaryConditions,
    ldg_bdryconds::Maybe{BoundaryConditions}, av_bdryconds::Maybe{BoundaryConditions})

  @unpack cache = rhs
  wrap_dynamic_variables!(cache, state_u)
  wrap_rhs_variables!(cache, state_du)
  rhs!(rhs, t, rhs.hrsc, bdryconds, ldg_bdryconds, av_bdryconds)
end


function rhs!(rhs::RHS, t, hrsc::Nothing, bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh, equation, rsolver = rhs
  @unpack D, S, tau             = get_dynamic_variables(cache)
  @unpack flx_D, flx_S, flx_tau = get_static_variables(cache)
  @unpack rhs_D, rhs_S, rhs_tau = get_rhs_variables(cache)
  @unpack lhs_numflx_D,   rhs_numflx_D,
          lhs_numflx_S,   rhs_numflx_S,
          lhs_numflx_tau, rhs_numflx_tau = get_bdry_variables(cache)

  broadcast_volume!(cons2prim, equation, cache)
  broadcast_volume!(flux, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
  broadcast_boundaryconditions!(lax_friedrich_flux, bdryconds, cache, mesh, t)

  compute_rhs_weak_form!(rhs_D,   flx_D,    lhs_numflx_D,   rhs_numflx_D,   mesh)
  compute_rhs_weak_form!(rhs_S,   flx_S,    lhs_numflx_S,   rhs_numflx_S,   mesh)
  compute_rhs_weak_form!(rhs_tau, flx_tau,  lhs_numflx_tau, rhs_numflx_tau, mesh)

  return
end


function rhs!(rhs::RHS, t, hrsc::HRSC.AbstractReconstruction,
    bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh, equation, rsolver = rhs
  @unpack D, S, tau                      = get_dynamic_variables(cache)
  @unpack flx_D, flx_S, flx_tau          = get_static_variables(cache)
  @unpack rhs_D, rhs_S, rhs_tau          = get_rhs_variables(cache)
  @unpack lhs_numflx_D,   rhs_numflx_D,
          lhs_numflx_S,   rhs_numflx_S,
          lhs_numflx_tau, rhs_numflx_tau = get_bdry_variables(cache)
  @unpack flag                           = get_cell_variables(cache)

  HRSC.reconstruct!(D,    flag, hrsc, isperiodic=isperiodic(bdryconds), m=1e-10, limit_with_boundaries=false)
  HRSC.reconstruct!(S,    flag, hrsc, isperiodic=isperiodic(bdryconds), limit_with_boundaries=false)
  HRSC.reconstruct!(tau,  flag, hrsc, isperiodic=isperiodic(bdryconds), m=1e-10, limit_with_boundaries=false)

  broadcast_volume!(cons2prim, equation, cache)
  broadcast_volume!(flux, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)

  compute_rhs_weak_form!(rhs_D,   flx_D,    lhs_numflx_D,   rhs_numflx_D,   mesh)
  compute_rhs_weak_form!(rhs_S,   flx_S,    lhs_numflx_S,   rhs_numflx_S,   mesh)
  compute_rhs_weak_form!(rhs_tau, flx_tau,  lhs_numflx_tau, rhs_numflx_tau, mesh)

  return
end


function rhs!(rhs::RHS, t, hrsc::HRSC.AbstractArtificialViscosity,
    bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh, equation, rsolver,
          ldg_rsolver, av_rsolver                = rhs
  @unpack D, S, tau                              = get_dynamic_variables(cache)
  @unpack flx_D, flx_S, flx_tau,
          ldg_D, ldg_S, ldg_tau,
          flx_ldg_D, flx_ldg_S, flx_ldg_tau      = get_static_variables(cache)
  @unpack rhs_D, rhs_S, rhs_tau                  = get_rhs_variables(cache)
  # TODO dispose of lhs/rhs numflx since we only need one value per interface
  # Does a penalty flux violate this assumption?
  @unpack lhs_numflx_D,   rhs_numflx_D,
          lhs_numflx_S,   rhs_numflx_S,
          lhs_numflx_tau, rhs_numflx_tau,
          lhs_numflx_ldg_D,   rhs_numflx_ldg_D,
          lhs_numflx_ldg_S,   rhs_numflx_ldg_S,
          lhs_numflx_ldg_tau, rhs_numflx_ldg_tau = get_bdry_variables(cache)

  ## solve auxiliary equation: q + ∂x u = 0
  # improved syntax
  # broadcast_volume!(ldg_flux,               equation,         grid)
  # broadcast_faces!(mean_flux,               ldg_rsolver,      grid)
  # broadcast_boundaryconditions!(mean_flux,  ldg_bdryconds(t), grid)
  broadcast_volume!(ldg_flux, equation, cache)
  broadcast_faces!(mean_flux, ldg_rsolver, cache, mesh)
  broadcast_boundaryconditions!(mean_flux, ldg_bdryconds, cache, mesh, t)
  compute_rhs_weak_form!(ldg_D,   flx_ldg_D,   lhs_numflx_ldg_D,   rhs_numflx_ldg_D,   mesh)
  compute_rhs_weak_form!(ldg_S,   flx_ldg_S,   lhs_numflx_ldg_S,   rhs_numflx_ldg_S,   mesh)
  compute_rhs_weak_form!(ldg_tau, flx_ldg_tau, lhs_numflx_ldg_tau, rhs_numflx_ldg_tau, mesh)

  ## compute rhs of regularized equation: ∂t u + ∂x f + ∂x μ q = 0
  broadcast_volume!(av_flux, equation, cache)
  broadcast_faces!(mean_flux, av_rsolver, cache, mesh)
  broadcast_boundaryconditions!(mean_flux, av_bdryconds, cache, mesh, t)

  broadcast_volume!(cons2prim, equation, cache)
  broadcast_volume!(flux, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
  broadcast_boundaryconditions!(lax_friedrich_flux, bdryconds, cache, mesh, t)

  # TODO Remove doubled memeory and calls below by allowing to broadcast with accumalative operations
  # e.g. broadcast_volume!([av_flux, flux], equation, cache)
  # which only works if av_flux and flux use the same outputs
  @. flx_D          += flx_ldg_D
  @. flx_S          += flx_ldg_S
  @. flx_tau        += flx_ldg_tau
  @. lhs_numflx_D   += lhs_numflx_ldg_D
  @. lhs_numflx_S   += lhs_numflx_ldg_S
  @. lhs_numflx_tau += lhs_numflx_ldg_tau
  @. rhs_numflx_D   += rhs_numflx_ldg_D
  @. rhs_numflx_S   += rhs_numflx_ldg_S
  @. rhs_numflx_tau += rhs_numflx_ldg_tau

  compute_rhs_weak_form!(rhs_D,   flx_D,   lhs_numflx_D,   rhs_numflx_D,   mesh)
  compute_rhs_weak_form!(rhs_S,   flx_S,   lhs_numflx_S,   rhs_numflx_S,   mesh)
  compute_rhs_weak_form!(rhs_tau, flx_tau, lhs_numflx_tau, rhs_numflx_tau, mesh)

  return
end



### Desired new interface
function rhs!(rhs::RHS, t, hrsc::HRSC.AbstractArtificialViscosity,
    bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack grid, equation, rsolver, ldg_rsolver, av_rsolver = rhs

  @unpack D, S, tau                      = dynamic_variables(cache)
  @unpack rhs_D, rhs_S, rhs_tau          = rhs_variables(cache)
  @unpack flx_D, flx_S, flx_tau,
          ldg_D, ldg_S, ldg_tau,         = static_variables(cache)
  @unpack numflx_D, numflx_S, numflx_tau = bdry_variables(cache)

  success = broadcast_volume!(cons2prim, equation, grid)

  ## solve auxiliary equation: q + ∂x u = 0
  # improved syntax
  broadcast_volume!(ldg_flux, equation, grid)
  broadcast_faces!(mean_flux, ldg_rsolver, grid)
  broadcast_boundaryconditions!(mean_flux, ldg_bdryconds(t), grid)

  compute_rhs_weak_form!(ldg_D,   flx_D,   numflx_D,   mesh)
  compute_rhs_weak_form!(ldg_S,   flx_S,   numflx_S,   mesh)
  compute_rhs_weak_form!(ldg_tau, flx_tau, numflx_tau, mesh)

  ## compute rhs of regularized equation: ∂t u + ∂x (f + μ q) = 0
  broadcast_volume!([av_flux, flux], equation, grid)
  broadcast_faces!([mean_flux, lax_friedrich_flux], [av_rsolver, rsolver], grid)
  broadcast_boundaryconditions!([mean_flux, lax_friedrich_flux],
                                [av_bdryconds(t), bdryconds(t)],
                                grid)

  compute_rhs_weak_form!(rhs_D,   flx_D,   numflx_D,   grid)
  compute_rhs_weak_form!(rhs_S,   flx_S,   numflx_S,   grid)
  compute_rhs_weak_form!(rhs_tau, flx_tau, numflx_tau, grid)

  return
end

### Other considerations
# - @with_signature: What is the maximum tuple size one can pass around?
#   If reached should be swap over to SVec?
#   Or are normal vectors good enough?
# - @with_signature: Think about @inline addition to macro.
