export RefVector, RefBlockVector, block, update!, eachblock


#######################################################################
#                              RefVector                              #
#######################################################################


"""
    RefVector{T} <: DenseVector{T}

A vector referencing a contiguous subrange of a `Vector{T}`.

---

    RefVector(v::Vector{T}[, offset, n])

Construct a `RefVector` tracking `v` starting at `offset` and length `n`.
Use `update!(refvec, vec[, offset, n])` to update the reference and/or subrange.

---

# Examples

```julia
julia> v = randn(100);

julia> refvec = RefVector(v, 10, 50);

julia> v[10] == refvec[1]
true

julia> length(refvec)
50

julia> update!(refvec, v, 20, 50);

julia> v[10] == refvec[1]
false
```
"""
mutable struct RefVector{T} <: DenseVector{T}
  ref::Base.RefValue{Vector{T}}
  # so far, vec is actually optional, but it is included for testing to see whether
  # the ref above can have any influence on performance
  vec::Vector{T}
  offset::Int64
  n::Int64

  function RefVector{T}(v::Vector{T}, offset::Integer, n::Integer) where T
    checkrange(length(v), offset, n)
    vec = unsafe_wrap(Vector{T}, pointer(v, offset), n)
    return new{T}(Ref(v), vec, offset, n)
  end
end

RefVector(v::Vector{T}, offset::Integer=1, n::Integer=length(v)) where T =
  RefVector{T}(v, offset, n)
RefVector(ref_v::Base.RefValue{Vector{T}}, offset::Integer=1, n::Integer=length(ref[])) where T =
  RefVector(ref_v[], offset, n)

# array interface
@inline Base.size(rv::RefVector) = (rv.n,)
@propagate_inbounds Base.getindex(rv::RefVector, i) = rv.vec[i]
@propagate_inbounds Base.setindex!(rv::RefVector, val, i) = rv.vec[i] = val
@inline Base.IndexStyle(A::RefVector) = IndexLinear()

Base.unsafe_convert(::Type{Ptr{T}}, rv::RefVector{T}) where T = pointer(rv.ref[], rv.offset)
# Base.unsafe_convert(::Type{Ptr{T}}, rv::RefVector{T}) where T = Base.unsafe_convert(Ptr{T}, rv.vec)

# @propagate_inbounds Base.getindex(rv::RefVector, i::Integer) = rv.ref[][i+rv.offset-1]
# @propagate_inbounds Base.setindex!(rv::RefVector, val, i::Integer) = rv.ref[][i+rv.offset-1] = val
# Base.iterate(rv::RefVector) = iterate(rv.vec)
# Base.iterate(rv::RefVector, state) = iterate(rv.vec, state)
# Base.BroadcastStyle(::Type{RefVector{T}}) where T = Base.BroadcastStyle(Vector{T})
# Base.similar(bc::Broadcasted{DestStyle}, ::Type{ElType})
# Base.copyto!(dest, bc::Broadcasted{DestStyle}) =
# function Base.show(io::IOBuffer, rv::RefVector)
#   println(io, "$(typeof(rv)):")
#   for v in rv
#       println(io, v)
#   end
#   return
# end


"""
    update!(rv::RefVector{T}, v::Vector{T} [, offset, n]) where T

Update the reference and/or subrange of `rv` to track `v`.
"""
function update!(rv::RefVector{T}, v::Vector{T}, offset=rv.offset, n=rv.n) where T
  checkrange(length(v), offset, n)
  rv.ref[] = v
  rv.vec = unsafe_wrap(Vector{T}, pointer(v, offset), n)
  rv.offset = offset
  rv.n = n
  return
end


function checkrange(N, offset, n)
  if !(1 <= offset <= N-n+1) || !(0 <= n <= N)
    error("Cannot reference vector of length $N starting from offset $offset with length $n")
  end
end


#######################################################################
#                           RefBlockVector                            #
#######################################################################


"""
		RefBlockVector{T,T_Vector<:DenseVector{T}} <: DenseVector{T}

A block vector referencing a `T_Vector<:DenseVector{T}`.

---

		RefBlockVector(v::Vector{T}, block_sizes[, offset])

Construct a `RefBlockVector` tracking `v` by blocks of length `block_sizes`,
optionally starting at `offset`.
Use `update!(refvec, vec[, block_sizes, offset])` to update the reference and/or blocks.

---

# Examples

```julia
julia> v = randn(100);

julia> refbvec = RefBlockVector(v, [10, 20, 30], 5)

julia> v[15:25] == block(refbvec, 1)
true

julia> length(refbvec)
60

julia> update!(refvec, v, [10, 20], 1);

julia> v[1:10] == block(refvec, 2)
true
```
"""
mutable struct RefBlockVector{T,T_Vector<:DenseVector{T}} <: DenseVector{T}
	ref::Base.RefValue{T_Vector}
	block_sizes::Vector{Int64}
	blocks::Vector{RefVector{T}}
	offset::Int64
  n::Int64

	function RefBlockVector{T,T_Vector}(v::T_Vector, block_sizes::Vector{Int64}, offset::Int64) where {T,T_Vector<:DenseVector{T}}
		blocks = RefVector{T}[]
		sizehint!(blocks, length(block_sizes))
		refvec = new{T,T_Vector}(Ref(v), block_sizes, blocks)
		update!(refvec, v, block_sizes, offset)
		return refvec
	end
end


RefBlockVector(v::T_Vector, block_sizes::Vector{Int64}, offset::Int64=1) where {T,T_Vector<:DenseVector{T}} =
	RefBlockVector{T,T_Vector}(v, block_sizes, offset)


"""
	update!(refvec::RefBlockVector{T}, v::T_Vector, [block_sizes, offset]) where {T,T_Vector<:DenseVector{T}}

Update the reference and/or blocks of `refvec` to track `v`.
"""
function update!(refvec::RefBlockVector{T}, v::T_Vector,
		block_sizes::Vector{Int64}=refvec.block_sizes, offset::Int64=refvec.offset) where {T,T_Vector<:DenseVector{T}}

	n = sum(block_sizes)
	check_block_sizes(length(v), n, block_sizes, offset)
	refvec.ref[]			 = v
	refvec.block_sizes .= block_sizes
	refvec.n           = n
	refvec.offset			 = offset
	block_offset       = offset
	resize!(refvec.blocks, length(block_sizes))
	for (i, bsz) in enumerate(block_sizes)
		if isassigned(refvec.blocks, i)
			update!(refvec.blocks[i], v, block_offset, bsz)
		else
			refvec.blocks[i] = RefVector{T}(v, block_offset, bsz)
		end
		block_offset += bsz
	end

	return
end


function check_block_sizes(N, n, block_sizes, offset)
	if !(1 <= offset <= N-n+1) || !(0 <= n <= N)
		error("Cannot reference vector of length $N starting from offset $offset with length $n")
	end
	if any(bsz -> bsz <= 0, block_sizes)
		error("block_sizes must be positive")
	end
end


# Abstract array interface
@inline Base.size(vec::RefBlockVector) = (vec.n,)
@propagate_inbounds Base.getindex(vec::RefBlockVector, i::Integer) = vec.ref[][i+vec.offset-1]
@propagate_inbounds Base.setindex!(vec::RefBlockVector, val, i::Integer) = vec.ref[][i+vec.offset-1] = val
@inline Base.IndexStyle(A::RefBlockVector) = IndexLinear()

# utilities
@propagate_inbounds block(vec::RefBlockVector, i) = vec.blocks[i]
eachblock(vec::RefBlockVector) = (block(vec, i) for i = 1:length(vec.blocks))
