export compute_rhs_weak_form!, 
       compute_rhs_strong_form!


"""
    compute_rhs_weak_form!(rhs, f, nf_lhs, nf_rhs, mesh)
    compute_rhs_weak_form!(rhs, f, s, nf_lhs, nf_rhs, mesh)
"""
function compute_rhs_weak_form!(rhs, f, nf_lhs, nf_rhs, mesh::Mesh)
  @unpack invjac, element = mesh
  @unpack invM, MDM, Ml_lhs, Ml_rhs, Npts = element
  @unpack K = mesh
  shape        = layout(mesh)
  mat_rhs      = reshape(view(rhs,:), shape)
  mat_f        = reshape(view(f,:), shape)
  # mat_MDM      = reshape(MDM, (Npts,Npts))
  # mul! currently allocates, although, (I think) it shouldn't ...
  # a = @allocated begin
  mul!(mat_rhs, MDM, mat_f)
  # end
  # display(a)
  @. mat_rhs  -= (nf_rhs' * Ml_rhs - nf_lhs' * Ml_lhs)
  mat_rhs    .*= invjac
  return
end
function compute_rhs_weak_form!(rhs, f, s, nf_lhs, nf_rhs, mesh::Mesh)
  compute_rhs_weak_form!(rhs, f, nf_lhs, nf_rhs, mesh)
  rhs .+= s
  return
end


"""
    compute_rhs_strong_form!(rhs, f, nf_lhs, nf_rhs, mesh)
    compute_rhs_strong_form!(rhs, f, s, nf_lhs, nf_rhs, mesh)
"""
function compute_rhs_strong_form!(rhs, f, nf_lhs, nf_rhs, mesh::Mesh)
  @unpack invjac, element = mesh
  @unpack Npts, invM, D, Ml_lhs, Ml_rhs = element
  @unpack K = mesh
  shape        = layout(mesh)
  mat_rhs      = reshape(view(rhs,:), shape)
  mat_f        = reshape(view(f,:), shape)
  # compute rhs = - D * f inplace to rhs
  # benchmarking showed that splitting of the factor of -1 is faster 
  # because a temporary is not created then
  mul!(mat_rhs, D, mat_f)
  mat_rhs     .*= -1
  @. mat_rhs  += (nf_rhs' * Ml_rhs - nf_lhs' * Ml_lhs)
  mat_rhs    .*= invjac
  return
end
function compute_rhs_strong_form!(rhs, f, s, nf_lhs, nf_rhs, mesh::Mesh)
  compute_rhs_strong_form!(rhs, f, nf_lhs, nf_rhs, mesh)
  rhs .+= s
  return
end


function compute_rhs_weak_form!(rhs, sf1, sf2, nsf1, nsf2, grid::Grid2d)

  for cell in grid.tree

    basis = grid.bases[cell.index]
    offset = grid.offsets[cell.index]

    vol_rng = range_volume(basis, offset)
    bdry_rng = range_boundary(basis, offset)

    @views begin
      vec_rhs = rhs[vol_idx]
      vec_f1, vec_f2 = f1[vol_rng], f2[vol_rng]
      # do we need four views here? I could imagine that two bigger ones suffice here to
      # if we make use of the conservation property of fluxes
      vec_nsf1, vec_nsf2 = nfs1[bdry_rng], nfs2[bdry_rng]

      cell_rhs = reshape(vec_rhs, vol_sz)
      cell_f1  = reshape(vec_f1, vol_sz)
      cell_f2  = reshape(vec_f2, vol_sz)
      # no reshaping needed because in 1D boundarys are lines
      cell_nsf1 = vec_nsf1
      cell_nsf2 = vec_nsf2
    end

    b1, b2 = basis
    nx, ny = size(basis)

    # add up derivatives
    muladd!(cell_rhs1, b1.D, cell_f1, -1.0, false)
    muladd!(transpose(cell_rhs1), cell_f2, b2.D, -1.0, 1.0)

    # add up numerical fluxes
    # xmin
    for (i, I) in enumerate(eachindex(@view(cell_rhs, 1, :)))
      mat_rhs[I] += (sub_nf1max[i] * b1.Ml_rhs - sub_nf1min[i] * b1.Ml_lhs)
    end
    # xmax
    for (i, I) in enumerate(eachindex(@view(cell_rhs, nx, :)))
      mat_rhs[I] += (sub_nf1max[i] * b1.Ml_rhs - sub_nf1min[i] * b1.Ml_lhs)
    end
    # ymin
    for (i, I) in enumerate(eachindex(@view(cell_rhs, :, 1)))
      mat_rhs[I] += (sub_nf2max[i] * b2.Ml_rhs - sub_nf2min[i] * b2.Ml_lhs)
    end
    # ymax
    for (i, I) in enumerate(eachindex(@view(cell_rhs, :, ny)))
      mat_rhs[I] += (sub_nf2max[i] * b2.Ml_rhs - sub_nf2min[i] * b2.Ml_lhs)
    end

  end
  # apply jacobian
  # TODO Benchmark whether this should be done here or inside the loop.
  mat_rhs .*= invjac
  return
end
function compute_rhs_weak_form!(rhs, s, sf1, sf2, nsf1, nsf2, grid::Grid2d)
  compute_rhs_weak_form!(rhs, sf1, sf2, nsf1, nsf2, grid)
  # an explicit source variable might not be needed, instead one could directly evaluate
  # the source terms into the rhs when computing them
  rhs .+= s
  return
end


function compute_rhs_weak_form!(rhs, sf1, sf2, sf3, nsf1, nsf2, nsf3, grid::Grid3d)

  for cell in grid.tree

    basis = grid.bases[cell.index]
    offset = grid.offsets[cell.index]

    vol_rng = range_volume(basis, offset)
    bdry_rng = range_boundary(basis, offset)

    @views begin
      vec_rhs = rhs[vol_idx]
      vec_f1, vec_f2, vec_f3 = f1[vol_rng], f2[vol_rng], f3[vol_rng]
      # do we need four views here? I could imagine that two bigger ones suffice here to
      # if we make use of the conservation property of fluxes
      vec_nsf1, vec_nsf2, vec_nsf3 = nfs1[bdry_rng], nfs2[bdry_rng], nsf3[bdry_rng]

      cell_rhs = reshape(vec_rhs, vol_sz)
      cell_f1  = reshape(vec_f1, vol_sz)
      cell_f2  = reshape(vec_f2, vol_sz)
      cell_f3  = reshape(vec_f3, vol_sz)
      # no reshaping needed because in 1D boundarys are lines
      cell_nsf1 = vec_nsf1
      cell_nsf2 = vec_nsf2
    end

    b1, b2 = basis
    nx, ny = size(basis)

    # add up derivatives
    muladd!(cell_rhs1, b1.D, cell_f1, -1.0, false)
    muladd!(transpose(cell_rhs1), cell_f2, b2.D, -1.0, 1.0)

    # add up numerical fluxes
    # xmin
    for (i, I) in enumerate(eachindex(@view(cell_rhs, 1, :)))
      mat_rhs[I] += (sub_nf1max[i] * b1.Ml_rhs - sub_nf1min[i] * b1.Ml_lhs)
    end
    # xmax
    for (i, I) in enumerate(eachindex(@view(cell_rhs, nx, :)))
      mat_rhs[I] += (sub_nf1max[i] * b1.Ml_rhs - sub_nf1min[i] * b1.Ml_lhs)
    end
    # ymin
    for (i, I) in enumerate(eachindex(@view(cell_rhs, :, 1)))
      mat_rhs[I] += (sub_nf2max[i] * b2.Ml_rhs - sub_nf2min[i] * b2.Ml_lhs)
    end
    # ymax
    for (i, I) in enumerate(eachindex(@view(cell_rhs, :, ny)))
      mat_rhs[I] += (sub_nf2max[i] * b2.Ml_rhs - sub_nf2min[i] * b2.Ml_lhs)
    end

  end
  # apply jacobian
  # TODO Benchmark whether this should be done here or inside the loop.
  mat_rhs .*= invjac
  return
end
function compute_rhs_weak_form!(rhs, s, sf1, sf2, sf3, nsf1, nsf2, nsf3, grid::Grid3d)
  compute_rhs_weak_form!(rhs, sf1, sf2, sf3, nsf1, nsf2, nsf3, grid)
  # an explicit source variable might not be needed, instead one could directly evaluate
  # the source terms into the rhs when computing them
  rhs .+= s
  return
end
